from redneck import config, resolver, builder

from argparse import Namespace
from pathlib import Path
import logging as log

def groups(args: Namespace):
    proj = config.scan_project(Path("."))
    if proj == None:
        return

    log.info(f"Enumerated {len(proj.decls)} mods")

    groups: dict[str, list[config.ModDecl]] = {}

    for id, decl in proj.decls.items():
        if decl.group not in groups:
            groups[decl.group] = []

        groups[decl.group].append(decl)

    log.info(f"Found {len(groups)} groups:")

    if args.full:
        for group, mods in groups.items():
            log.info(f"- {group}:")
            for mod in mods:
                log.info(f"  * {mod.id} ({mod._from_file}:{mod._line}:{mod._col})")
    else:
        for group, mods in groups.items():
            log.info(f"- {group}")


from argparse import Namespace
from redneck import config, resolver, builder

import logging as log
from pathlib import Path

def build(args: Namespace):
    if args.builder == "list":
        return

    groups_split: set[str] = set()
    if args.groups != None:
        groups_split = {
            group.strip()
            for group in str(args.groups).split(",")
            if group.strip()
        }
        log.debug(f"Also using group(s) {sorted(groups_split)}")

    proj = config.scan_project(Path("."))
    if proj == None:
        return    

    used_mods: list[config.ModDecl] = []

    for id, decl in proj.decls.items():
        if decl.group == "base" or decl.group in groups_split:
            used_mods.append(decl)

    log.info(f"Enumerated {len(proj.decls)} mods, resolving {len(used_mods)} mods")

    mods = resolver.resolve_mods(used_mods)
    project = builder.ResolvedProject(proj.meta, mods)

    if args.builder == "all":
        for b in list(builder._builders.keys()):
            final_build = builder.build_project(b, project)

            if final_build == None:
                log.error("Build failed")
                return

            log.info(f"Built {final_build}")
    else:
        final_build = builder.build_project(args.builder, project)

        if final_build == None:
            log.error("Build failed")
            return

        log.info(f"Built {final_build}")



from csv import Error
import logging as log
from typing import Generic, TypeVar

from redneck.config import ModDecl, RedneckProject

T = TypeVar("T", bound=ModDecl)

class ResolvedMod:
    url: str
    extra_info: dict[str, str]

    def __init__(self, url: str, extra_info: dict[str, str]) -> None:
        self.url = url
        self.extra_info = extra_info

class ModResolver(Generic[T]):
    def resolve(self, decl: T) -> ResolvedMod:
        raise NotImplementedError
    def health_check(self, proj: RedneckProject, decl: T) -> list[str]:
        raise NotImplementedError

def resolve_mods(proj: list[ModDecl]) -> list[ResolvedMod]:
    mods = []
    for decl in proj:
        try:
            mod = _resolvers[decl.load].resolve(decl)
        except Error as e:
            log.error(f"Failed to resolve {decl.id}: {e}")

        mod.extra_info["resolver"] = decl.load
 
        mods.append(mod)


    return mods

def health_check(proj: RedneckProject) -> dict[str, list[str]]:
    warnings = {}

    for decl in list(proj.decls.values()):
        warns = _resolvers[decl.load].health_check(proj, decl)

        if len(warns) == 0:
            continue

        warnings[decl.id] = warns

    return warnings

_resolvers: dict[str, ModResolver[ModDecl]] = {}


import logging as log

from argparse import Namespace
from pathlib import Path


def init(args: Namespace):
    root = Path(".")
    mods = (root / "mods/")

    if not mods.exists():
        mods.mkdir()

    with open(root / "pack.yml", "w") as f:
        f.write("""pack:
  id: "cool-pack"
  name: "Cool pack"
  version: "1.0.0"

versions:
  minecraft: "1.20.1"
  fabric-loader: "0.18.4\"""")

    with open(mods / "hello.yml", "w") as f:
        f.write("""- load: modrinth
  id: fabric-api
  project: "fabric-api"
  version: "0.92.7+1.20.1\"""")

    log.info(f"Initialized empty redneck project in {root.absolute()}")
    log.info(f"Hint: try building the project with \"redneck build mrpack\"")


import logging as log
from functools import reduce
from pathlib import Path
from typing import Annotated, Any, ClassVar
from yaml import Loader, load, SafeLoader
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, TypeAdapter, ValidationError

class ModDeclLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = super(ModDeclLoader, self).construct_mapping(node, deep=deep)
        mapping["_line"] = node.start_mark.line + 1
        mapping["_col"] = node.start_mark.column + 1

        return mapping

class ModpackMeta(BaseModel):
    class PackInfo(BaseModel):
        id: str
        name: str
        version: str

    pack: PackInfo
    minecraft: str
    loader: str
    loader_version: str
    builders: dict[str, dict[str, Any]] = {}

class ModDecl(BaseModel):
    id: str
    group: str = "base"

    _line: int = PrivateAttr(default=-1)
    _col: int = PrivateAttr(default=-1)
    _from_file: str = PrivateAttr(default="unknown")
    model_config: ClassVar[ConfigDict] = ConfigDict(extra='forbid')

class RedneckProject:
    meta: ModpackMeta
    decls: dict[str, ModDecl]

    def __init__(self, meta: ModpackMeta, decls: dict[str, ModDecl]) -> None:
        self.meta = meta
        self.decls = decls

_decls: list[type[ModDecl]] = []

def scan_project(root: Path) -> RedneckProject | None:
    if not (root / "pack.yml").exists() or len(list((root / "mods/").glob("*.yml"))) == 0:
        log.error("Missing either pack.yml, the \"mods/\" dir, or the yml files in mods dir.")
        return None

    union_type = reduce(lambda a, b: a | b, _decls)
    mod_adapter = TypeAdapter(Annotated[union_type, Field(discriminator="load")])

    mods = {}

    try:
        with open(root / "pack.yml", 'r') as stream:
            meta = load(stream, Loader=SafeLoader)

            # Bodge to load the loader & loader name from yaml.
            # I couldn't find a better way to do this, sorry.

            versions = meta.pop("versions", {})
            loaders = ["fabric-loader", "forge", "quilt-loader", "neoforge"]
            minecraft = versions.get("minecraft")
            
            loader = next((i for i in versions if i in loaders), None)
    
            if not minecraft:
                raise Exception("Missing \"minecraft\" in the \"versions\" block.", )
            if not loader:
                raise Exception(f"Missing a loader in the \"versions\" block. Declare one of: {", ".join(loaders)}.")

            meta["minecraft"] = minecraft
            meta["loader"] = loader
            meta["loader_version"] = versions[loader]

        for file in list((root / 'mods').glob("*.yml")):
            with open(file, 'r') as stream:
                log.debug(f"Loading {file}")
                loaded = load(stream, Loader=ModDeclLoader)

                if loaded == None:
                    log.warning(f"In {file}: file is empty!")
                    continue

                for decl in loaded:
                    id = decl["id"]

                    line = decl.pop("_line", None)
                    col = decl.pop("_col", None)

                    mod = mod_adapter.validate_python(decl)
                    mod._from_file = file
                    mod._line = line
                    mod._col = col

                    if id in mods:
                        log.warning(f"In {file}:{mod._line}:{mod._col}: {id} already defined!")
                        log.warning(f"First seen in: {mods[id]._from_file}:{mods[id]._line}:{mods[id]._col}")

                    mods[id] = mod

        return RedneckProject(ModpackMeta(**meta), mods)
    except Exception as e:
        log.error(f"Project scan failure:\n{e}")